<?php

namespace NoccyLabs\React\Daemon;

use Evenement\EventEmitterInterface;
use NoccyLabs\React\Daemon\Attributes\Help;
use NoccyLabs\React\Daemon\Attributes\OnEvent;
use NoccyLabs\React\Daemon\Attributes\Option;
use React\EventLoop\Loop;
use ReflectionObject;

abstract class Daemon
{
    private $_startCb, $_stopCb;

    private array $_options = [];

    private array $events = [];

    private array $timers = [];

    private array $intervals = [];

    private array $modules = [];

    private array $options = [];

    private mixed $entry = null;

    private ?Help $help = null;

    private bool $stopping = false;

    public function __construct()
    {
    }

    public function help(): void
    {
        ob_start();

        if ($this->help) {
            printf("\e[1m%s\e[0m\n\e[3m%s\e[0m\n\n", $this->help->name, $this->help->summary);
        }
        printf("Options:\n");
        foreach ($this->options as [$option,$prop]) {
            $opt = array_filter([ $option->short ? "-{$option->short}" : "", $option->long ? "--{$option->long}" : "" ]);
            $opt = join(",",$opt);
            printf(" \e[32m%-20s \e[33m%s\e[0m", $opt, $option->help);
            if ($option->default) {
                printf(" (\e[3m%s\e[0m)", $option->default);
            } elseif ($prop->getDefaultValue()) {
                printf(" (\e[3m%s\e[0m)", $prop->getDefaultValue());
            }
            if ($option->env) {
                printf("   \e[36m\$%s\e[0m", $option->env);
            }
            echo "\n";
        }

        $out = ob_get_contents();
        ob_end_clean();

        if (!posix_isatty(STDOUT)) {
            $out = preg_replace("/(\e\[(.+?)m)/m", "", $out);
        }
        echo $out;
    }

    final public static function boot(Daemon $daemon): Daemon
    {
        //echo "Boot\n";
        $daemon->loadAttributes();
        $daemon->parseOptions();
        if (is_callable($daemon->entry)) {
            call_user_func($daemon->entry);
        } else {
            $daemon->start();
        }
        return $daemon;
    }

    private function loadAttributes(): void
    {
        $r = new ReflectionObject($this);

        $help = $r->getAttributes(Help::class);
        if ($help) $this->help = reset($help)->newInstance();

        foreach ($r->getProperties() as $property) {
            $options = $property->getAttributes(Option::class);
            foreach ($options as $option) {
                $option = $option->newInstance();
                $this->options[] = [ $option, $property ];
            }
        }

        foreach ($r->getMethods() as $method) {
            $events = $method->getAttributes(OnEvent::class);
            foreach ($events as $event) {
                $event = $event->newInstance();
                $eventName = $event->event;
                if (!isset($this->events[$eventName])) {
                    $this->events[$eventName] = [];
                }
                $this->events[$eventName][] = $method->getClosure($this);
                // if ($eventName == '@start') {
                //     $this->_startCb = $method->getClosure($this);
                // }
                // if ($eventName == '@stop') {
                //     $this->_stopCb = $method->getClosure($this);
                // }
            }
        }
    }

    final protected function hasEvent(string $event): bool
    {
        return isset($this->events[$event]);
    }

    final public function invokeEvent(string $name, mixed $data=null): void
    {
        $events = $this->events[$name] ?? null;
        if ($events) {
            foreach ($events as $event) {
                call_user_func($event, $data);
            }
        }
        foreach ($this->modules as $module) {
            $module->onEvent($name, $data);
        }
    }

    private function parseOptions(): void
    {
        $longOpts = [];
        $shortOpts = [];
        foreach ($this->options as [$option,$property]) {
            /** @var Option $option */
            /** @var ReflectionProperty $property */
            $typeref = $property->getType();
            $type = (string)$typeref;
            $args = ($type == 'bool')?"":":";
            if ($option->long) $longOpts[] = $option->long . $args;
            if ($option->short) $shortOpts[] = $option->short . $args;
        }
        $parsed = getopt(join("",$shortOpts), $longOpts, $optind);
        foreach ($this->options as [$option,$property]) {
            /** @var Option $option */
            /** @var ReflectionProperty $property */
            $typeref = $property->getType();
            $type = (string)$typeref;
            $property->setAccessible(true);
            if ($option->long && array_key_exists($option->long, $parsed)) {
                if ($type == 'bool') {
                    $value = true;
                } else {
                    $value = $parsed[$option->long];
                }
                $property->setValue($this, $value);
                if ($option->entrypoint) {
                    $this->entry = $this->{$option->entrypoint}(...);
                }
            } elseif ($option->short && array_key_exists($option->short, $parsed)) {
                if ($type == 'bool') {
                    $value = true;
                } else {
                    $value = $parsed[$option->short];
                }
                $property->setValue($this, $value);
                if ($option->entrypoint) {
                    $this->entry = $this->{$option->entrypoint}(...);
                }    
            } elseif ($option->env && $env = getenv($option->env)) {
                $property->setValue($this, $env);
            } elseif ($option->default !== null) {
                $property->setValue($this, $option->default);
            }
        }
    }

    final public function start(): void
    {
        //if (is_callable($this->entry)) {
        //    call_user_func($this->entry);
        //    return;
        //}

        if ($this->hasEvent('@start')) {
            $this->invokeEvent('@start');
        }
        // if (is_callable($this->_startCb)) {
        //     call_user_func($this->_startCb);
        // }
        pcntl_signal(SIGHUP, $this->onSignal(...));
        pcntl_signal(SIGINT, $this->onSignal(...));
        pcntl_signal(SIGQUIT, $this->onSignal(...));
        pcntl_signal(SIGTERM, $this->onSignal(...));
        pcntl_async_signals(true);
    }

    final public function stop(bool $force = false): void
    {
        if ($this->hasEvent('@stop')) {
            $this->invokeEvent('@stop');
        }
        // if (is_callable($this->_stopCb)) {
        //     call_user_func($this->_stopCb);
        // }
        $this->clearAll();
        $this->unloadAllModules();

        if ($this->stopping) {
            exit(255);
        } else {
            $this->stopping = true;
        }
    }

    private function onSignal(int $signo): void
    {
        switch ($signo) {
            case SIGHUP:
                if ($this->hasEvent("@reload")) {
                    $this->invokeEvent("@reload");
                } else {
                    $this->stop();
                }
                break;
            case SIGINT:
            case SIGQUIT:
                $this->stop();
                break;
            case SIGTERM:
                exit(255);
        }
    }

    final public function addTimer(int|float $interval, callable $callback): void
    {
    }

    final public function addNamedTimer(string $name, int|float $interval, callable $callback): void
    {
        $this->timers[$name] = Loop::addTimer($interval, $callback);
    }

    final public function clearNamedTimer(string $name): void
    {
        $timer = $this->timers[$name]??null;
        if (!$timer) return;
        unset($this->timers[$name]);
        Loop::cancelTimer($timer);
    }

    final public function addNamedInterval(string $name, int|float $interval, callable $callback): void
    {
        $this->intervals[$name] = Loop::addPeriodicTimer($interval, $callback);
    }

    final public function clearNamedInterval(string $name): void
    {
        $interval = $this->intervals[$name]??null;
        if (!$interval) return;
        unset($this->intervals[$name]);
        Loop::cancelTimer($interval);
    }

    final public function clearAll(): void
    {
        foreach (array_keys($this->timers) as $name) {
            $this->clearNamedTimer($name);
        }
        foreach (array_keys($this->intervals) as $name) {
            $this->clearNamedInterval($name);
        }
    }

    final public function module(string $name): ?ModuleInterface
    {
        return $this->modules[$name]??null;
    }

    final public function loadModule(string $name, ModuleInterface $module): void
    {
        $this->modules[$name] = $module;
        if ($module instanceof EventEmitterInterface) {
            $module->on('event', function (array $arg) {
                $this->invokeEvent($arg[0], $arg[1]??[]);
            });
        }
        $module->onLoad();
    }

    final public function unloadModule(string $name): void
    {
        $module = $this->modules[$name]??null;
        if (!$module) return;
        if ($module instanceof EventEmitterInterface) {
            $module->removeAllListeners();
        }
        unset($this->modules[$name]);
        $module->onUnload();
    }

    final public function unloadAllModules(): void
    {
        while ($module = array_shift($this->modules)) {
            $module->onUnload();
            unset($module);
        }
    }

}

